;; pathfinder egg
;; Author: Jim Ursetto, Ursetto Consulting, Inc.
;; See LICENSE for copyright

;; Acknowledgements: Some design inspiration as well as handling of compound
;;    extensions was taken from the MIT-licensed Hike library by
;;    Sam Stephenson, https://github.com/sstephenson/hike

(module pathfinder
  (make-pathfinder
   pathfinder-paths pathfinder-root current-pathfinder
   path-find path-find-all path-fold
   pathfinder-reset
   pathfinder-stat                ;; export not yet justified
   pathfinder-default-test
   pathfinder-default-matcher
   pf:exact pf:extensions pf:compound
   pf:regular-file? pf:directory?
   dirent-exists? dirent-contents dirent-stat dirent-directory
   )

(import scheme chicken)
(use posix srfi-69 files)
(use (only posix-extras stat-regular-file? stat-directory?))  ;; for pf:xxx? tests
(use srfi-1)
(use data-structures)
(use regex)
(import irregex)
(import (prefix (only data-structures sort) core:))  ;; due to conflicting keywords
(import (prefix (only srfi-1 filter) core:))

;; Notes: symlinks will trick pathfinder into caching duplicate copies, as entries are keyed to full path
;;        and directories are not traversed individually to canonicalize the path.  Symlinks may also
;;        cause lookups to escape the root.
;; Notes: Probably not thread-safe to use same pathfinder object in multiple threads.
;; Notes: Pathnames must be normalized after make-pathname to correct the separator, as make-pathname
;;        always uses /, contrary to the docs.
;; FIXME: Normalize extensions by ensuring they contain start with a dot (and contain only one dot?)
;; TODO:  Also possible we could search for regex if we pass regex object in, and exact if we pass exact in.
;;        Would further eliminate need for separate searchers.  However, it does not eliminate the need for
;;        the compound matcher, because that does special sorting.
;; TODO:  make-pathfinder filter and sort are not yet finalized (and thus not documented).

(define-record-type pathfinder
  (%make-pathfinder paths entries root filter sorter searcher)
  pathfinder?
  (paths %pathfinder-paths)          ;; list: ordered search path of absolute pathnames
  (entries pathfinder-entry-cache pathfinder-entry-cache-set!)   ;; ht: "/path/to/dir/" => dirent ht
  (root %pathfinder-root)            ;; str: absolute root pathname
  (filter pathfinder-filter)         ;; (proc paths): path list => filtered path list (called when dirent first read)
  (sorter pathfinder-sorter)         ;; (proc paths): path list => sorted path list (called when dirent first read)
  (searcher pathfinder-searcher)     ;; 
  )

;; dirent ht
(define-record-type dirent
  (%make-dirent table list dir)
  dirent?
  (table dirent-table) (list dirent-list) (dir dirent-directory))

(define (make-dirent dir basenames)    ;; primary constructor
  (define (alist->dirent a)
    (alist->hash-table a test: string=?))
  (%make-dirent (alist->dirent
                 (map (lambda (x) (cons x #t)) ;; NB: May be able to store stat result here.
                      basenames))
                basenames
                dir))
(define (dirent-ref de basename)          ;; Internal use for getting stat vector.
  (hash-table-ref/default (dirent-table de) basename #f))
(define (dirent-exists? de fn)            ;; User visible boolean variant of dirent-ref.
  (and (dirent-ref de fn) #t))
(define (dirent-set! de fn s)
  (hash-table-set! (dirent-table de) fn s))
(define dirent-contents dirent-list)

;; Stat file represented by FN in dirent DE and return stat vector.  Stat data is cached at first lookup.
;; Note: if the stat fails, #f is returned and #f is stored in the cache; the entry remains in the dirent
;; but it is subsequently treated as a missing file.
(define (dirent-stat de fn)
  (and-let* ((s (dirent-ref de fn)))
    (if (vector? s)
        s
        (let ((s (handle-exceptions e #f     ;; Gosh darn it we need a non-erroring file-stat
                   (file-stat (make-pathname ;; Separator wrong on windows, but should stat ok
                               (dirent-directory de) fn))))) ;; follow symlink correct?
          (dirent-set! de fn s)
          s))))

;; entry cache ht
(define (make-entry-cache) (make-hash-table string=?))
(define (entry-cache-set! ec path dirent)
  (hash-table-set! ec path dirent))
(define (entry-cache-ref ec path)
  (hash-table-ref/default ec path #f))

;; Fresh copies for user so mutation does not violate cache invariant
(define (pathfinder-paths pf) (list-copy (%pathfinder-paths pf)))
(define (pathfinder-root pf)  (string-copy (%pathfinder-root pf)))

;; Stat full pathname PN (possibly via stat cache).  Intended to be used on full pathname AFTER
;; the pathname has been returned from a successful search, e.g. inside a path-fold.  This does
;; not do a search itself; in particular, search paths which have not yet been scanned do not
;; have dirents and will return #f.  Semantics and rationale for this are iffy.
(define (pathfinder-stat pf pn)
  (define (normalize-dirname d)  ;; separator is appended
    (normalize-pathname (make-pathname d "")))
  (and-let* ((de (entry-cache-ref (pathfinder-entry-cache pf)
                                  (normalize-dirname (pathname-directory pn)))))
    (dirent-stat de (pathname-strip-directory pn))))

;;; pathfinder matchers

(define (pf:compound exts #!optional (test (pathfinder-default-test)))
  (define (make-extension-predicate exts)
    (let ((rx (irregex `(: bos (+ (or ,@exts)) eos))))
      (lambda (path term)       ;; FIXME: should pass dir, base, term?  but stat will probably take full path?
        ;; nb if we could get index of basename, strip is unnecessary; no known portable way in chicken though
        (let ((base (pathname-strip-directory path)))
          ;; FIXME!!! Parent already checked for exact match and failed.  Probably no need to repeat it
          (and (substring=? term base)   ;; does this go here?  should this just check extensions?
               (or (= (string-length term) (string-length base))
                   (string-search rx
                                  base
                                  (string-length term))))))))
  ;; sort 'seq' by expensive operation 'monetize' (via "schwartzian transform")
  (define (sort-by seq less? monetize)
    (map car (sort (map (lambda (x) (cons x (monetize x))) seq)
                   (lambda (x y) (less? (cdr x) (cdr y))))))
  (define sort-by-compound-extension
    (let ((+rx:exts+ (irregex "\\.[^.]+"))
          (priority (lambda (ext exts) (+ 1 (list-index (lambda (x) (string=? ext x)) exts)))))
      (lambda (exts basename files)
        (sort-by files <
                 (lambda (f)
                   (fold (lambda (x xs) (+ (priority x exts) xs)) 0
                         (string-split-fields +rx:exts+ f #t
                                              (string-length basename))))))))
  (let ((ext? (make-extension-predicate exts)))
    (lambda (de basename multiple?)
      (define (sort L)
        (sort-by-compound-extension exts basename L))
      (if multiple?
          (sort (filter (lambda (x) (and (ext? x basename)
                                    (or (not test) (test de x))))
                        (dirent-contents de)))
          (or (and (dirent-exists? de basename)
                   (or (not test) (test de basename))
                   basename)
              (let ((files (sort (filter (lambda (x) (and (ext? x basename)
                                                     (or (not test) (test de x))))
                                         (dirent-contents de)))))
                (and (pair? files) (car files)))
              ;; Cheaper way, if we don't have to sort every time.
              ;; (find (lambda (x) (ext? x basename))
              ;;       (dirent-contents de))
              )))))

(define (pf:exact #!optional (test (pathfinder-default-test)))
  (lambda (de fn all?)
    (if all?
        (if (and (dirent-exists? de fn) (or (not test) (test de fn)))
            (list fn) '())
        (if (and (dirent-exists? de fn) (or (not test) (test de fn)))
            fn #f))))

(define (pf:extensions exts #!optional (test (pathfinder-default-test)))
  (lambda (de fn all?)
    ;; Currently assume that exts are prefixed with periods, and we just append them in order.
    ;; If extension *matches* what is currently on file then we do NOT append it; i.e. "a" or "a.exe" will find
    ;; "a.exe" if extensions list is '(".exe").
    (define (match-ext e)
      (if (substring=? fn e (- (string-length fn) (string-length e))
                       0 (string-length e))
          (and (dirent-exists? de fn) fn) ;; Exact match when .ext is already on there.
          (let ((fe (string-append fn e)))
            (and (dirent-exists? de fe)
                 fe))))
    (define (test-ext e)
      (if test
          (let ((fn (match-ext e)))
            (and fn (test de fn) fn))
          (match-ext e)))
    (if all?
        (filter-map test-ext exts)
        (any test-ext exts))))

;;; pathfinder tests

(define (pf:regular-file? de fn)
  (and-let* ((s (dirent-stat de fn)))
    (stat-regular-file? s)))
(define (pf:directory? de fn)
  (and-let* ((s (dirent-stat de fn)))
    (stat-directory? s)))

;;; parameters

(define pathfinder-default-test (make-parameter pf:regular-file?))
(define pathfinder-default-matcher (make-parameter (pf:exact)))

;;; constructors

;; Create pathfinder object with search path PATHS (a list).  Search paths may be absolute or relative.
;; - Keys: root: Any relative pathnames in PATHS are taken as relative to ROOT.
;;               If ROOT itself is relative, it is taken as relative to the current working directory
;;               at the time of initialization of this object.
;;         filter: Procedure of 1 argument called with each BASENAME at dirent read time.  Note: this should
;;                 really take the dirent as well, but it hasn't been constructed yet.
;;         sort: Sort comparison (i.e. less?) procedure called with each BASENAME at dirent read time
(define (make-pathfinder paths #!key (matcher (pathfinder-default-matcher)) (root ".") (filter #f) (sort #f))
  (define (normalize-root r)
    (let ((r (normalize-pathname r)))
      (if (absolute-pathname? r)
          (string-copy r)
          (normalize-pathname (make-pathname (current-directory) r)))))
  (define (normalize-dirname d)  ;; appends separator
    (normalize-pathname (make-pathname d "")))
  (let ((root (normalize-root root)))
    (define (relative-to-root p)
      (let ((p (normalize-pathname p)))    ;; e.g. if p contains a ~
        (if (absolute-pathname? p)
            (normalize-dirname p)
            (normalize-dirname (make-pathname root p)))))
    (%make-pathfinder (map relative-to-root paths)
                      (make-entry-cache)
                      root
                      (if filter
                          (lambda (paths) (core:filter filter paths))
                          identity)
                      (if sort
                          (lambda (paths) (core:sort paths sort))
                          identity)
                      matcher)))

#|
;; Unused for now.
(define (compound-extension-pathfinder paths extensions #!key root filter)
  (make-pathfinder paths matcher: (pf:compound extensions)
                   root: root filter: filter sort: string<?))

(define (exact-pathfinder paths #!key root filter)  ;; sort irrelevant here
  (make-pathfinder paths matcher: (pf:exact)
                   root: root filter: filter))
|#

(define current-pathfinder (make-parameter #f))   ;; Used inside filters and sorters.  May be pointless.

;; Sequentially search for PATHNAME using pathfinder PF.
;; - If pathname is not absolute, it is taken relative to the paths in pathfinder-paths;
;;   each path is searched in order.
;; - If pathname is absolute, it is taken relative to ROOT; if this resulting path
;; is in the search path, do a lookup in that single path.  (If not in search path, return #f.)
;; - Returns the absolute path to the resource or #f if not found.
;; - As a side effect, any paths that are not in PF's entry cache are cached as they are encountered.
;; - NB Effect of using PATHNAME containing . or .. is officially undefined.  (Unofficially, all
;;   paths are normalized with normalize-pathname so . is removed and .. removes itself and its parent dir.)
;;   Using .. may cause repeated queries to the same normalized path.

(define (%pathfind pf pathname search foldfunc init)
  (let* ((basename (pathname-strip-directory pathname))
         (ec (pathfinder-entry-cache pf))
         (search (or search (pathfinder-searcher pf)))
         (filter-paths (pathfinder-filter pf))
         (sort-paths (pathfinder-sorter pf)))
    (define (find-in-one-path p multiple?)
      (define (search-dirent de)
        (and (dirent? de) ;; Skip if not dirent.
             (let ((contents (search de basename multiple?)))
               (if multiple?   ;; Eh.  Maybe search should always return a list.
                   (map (lambda (b) (string-append p b)) contents)      ;; note: p === (dirent-directory de)
                   (and contents (string-append p contents))))))   ;; note: p guaranteed to end in separator
      (or (cond ((entry-cache-ref ec p) => search-dirent)
                ((not (directory? p))
                 (entry-cache-set! ec p #t) ;; Record non-dirent for this path so we can skip this later.
                 #f)
                (else
                 (let ((de (make-dirent p (sort-paths (filter-paths (directory p 'dotfiles))))))
                   (entry-cache-set! ec p de)
                   (search-dirent de))))
          (and multiple? '())))             ;; In multiple search mode callee expects '() return, not #f
    (define (make-normalized-dir d b)  ;; append separator then normalize
      (normalize-pathname
       (let ((p (make-pathname d b)))
         (if (string=? b "") p (make-pathname p "")))))
    (define (subdir? d/ paths)  ;; #t if d/ is equal or subdir of any p/ in paths.
      ;; Assumes d/ and all paths are already normalized and also end in separator char.
      ;; (Appended separator char avoids false + if p=/foo/bar,d=/foo/bar1.)
      (any (lambda (p/)
             (and (>= (string-length d/) (string-length p/))
                  (substring=? d/ p/)))
           paths))

    (parameterize ((current-pathfinder pf))                 ;; May not be useful.
      (let ((dir (or (pathname-directory pathname) ""))
            (paths (%pathfinder-paths pf)))
        (if (absolute-pathname? dir)
            (let ((p (make-normalized-dir (pathfinder-root pf) dir)))
              (and (subdir? p paths)
                   (if foldfunc
                       (fold foldfunc init (find-in-one-path p #t))   ;; note: for map, incurs list copy and reverse!
                       (find-in-one-path p #f))))
            (if foldfunc
                (let outer ((p paths)
                            (s init))
                  (if (null? p)
                      s
                      (let inner ((f (find-in-one-path (make-normalized-dir (car p) dir) #t))
                                  (s s))
                        (if (null? f)
                            (outer (cdr p) s)
                            (inner (cdr f) (foldfunc (car f) s))))))
                (any (lambda (p) (find-in-one-path (make-normalized-dir p dir) #f))
                     paths)))))))

(define (path-find pf pathname #!optional search)
  (%pathfind pf pathname search #f #f))

(define (path-find-all pf pathname #!optional search)
  (reverse (%pathfind pf pathname search cons '())))

;; It's easier to use filter + path-find-all than to do filtering inside the fold.
;; Also no way to escape fold w/o call/cc.
(define (path-fold pf func init pathname #!optional search)
  (%pathfind pf pathname search func init))

;; Reset cached data for pathfinder PF.
(define (pathfinder-reset pf)
  (pathfinder-entry-cache-set! pf (make-entry-cache))
  (void))


)

